# -*- coding: utf-8 -*-
"""
Created on Thu Jan 16 13:43:08 2020

General description: 
-------------------------------------------------------------------------------
This model is a basic model for energysystem analysis following a linear
optimization and regarding future technologies like Power-to-Hydrogen.
It is based on the open-source Modelling framework Oemof.

Documentation:
-------------------------------------------------------------------------------
The development and set up of the model is described in the Master thesis final document.
All relevant files for the model run can be found in the electronical appendix A 3.
Besides inline comments are implemented in the code.
Specific further information regarding oemoef components can be found online.
URL: https://oemof.readthedocs.io/en/stable/using_oemof.html


Additional comments:
-------------------------------------------------------------------------------
Inline comments and code are written in English. Labels, that are used later for
the results anaysis, are written in German to facilitate the visualization of
the results in the final thesis document.


Input Data:
-------------------------------------------------------------------------------
Anhang A.3/input_data_g1
Anhang A.3/input_data_g2
Anhang A.3/input_data_g3
Anhang A.3/input_data_g4
Anhang A.3/input_data_g5

Anhang A.3.1 contains the input data sheets. For the model run the data files must 
be located in the same folder like the main model file.

Output files:
-------------------------------------------------------------------------------
Anhang A.3/Results_Wismar
Anhang A.3/Results_Grevesmühlen
Anhang A.3/Results_Selmsdorf
Anhang A.3/Results_Schildetal
Anhang A.3/Results_Zurow

Anhang A.3 contains the Excel file, where the model results are exported to.
For the Model run, the result files must be located in the same folder like
the main model file.


Imported files:
-------------------------------------------------------------------------------
VisioData_Auto_v01 is just used for defining the graph for plotting


Installation requirements:
-------------------------------------------------------------------------------
Python version >= 3.7.6
oemof version 0.3.2

@author: Charlotte Jarosch
Thanks goes to the oemof developer group and furthermore to
the department Energy and Resources of TU Berlin, who supported the development.
"""

##### Importing required packages #############################################

import os
import logging
import pandas as pd
import numpy as np 
import pyomo.environ as po

from oemof.tools import logger
from oemof import solph
from oemof import outputlib
from oemof.tools import economics

from matplotlib import pyplot as plt

from collections import OrderedDict
#import oemof_visio as oev
from openpyxl import load_workbook

import VisioData_auto_v01


###############################################################################
### Model Configuration
###############################################################################

# Export Output to an excel table (True/False):
export_output = True

# Save the plots (True/False): If True, a folder results is needed to save the plots:
save_plots = False

# Calculate the dual variables (True/False): 
obtain_duals = True

# Defiine the scenario: Status quo = 0 , Zukunft = 50
scenario_id = 0

# Define the modelled energysystem: Gemeinde_ID: 1: Wismar, 2: Grevesmühlen, 3: Selmsdorf, 4: Schildetal, 5:Zurow
gemeinde_id = 5

# Store the scenario description in a dictionary:
scenario= {'0':'Status quo', '50': 'Zukunft', '1': 'Wismar', '2':'Grevesmühlen', '3':'Selmsdorf',
           '4': 'Schildetal', '5':'Zurow'}

# Investment mode for the PtH2 components (True/False):
invest_mode = True

# Function: Choose input-excel depending on chosen municipilty 
def define_gemeinde(gemeinde_id):
    if gemeinde_id== 1:
        filename= 'input_data_g'+str(gemeinde_id)+'.xlsx'
        print('Gemeinde: '+ scenario[str(gemeinde_id)])
    
    if gemeinde_id == 2: 
        filename= 'input_data_g'+str(gemeinde_id)+'.xlsx'
        print('Gemeinde: '+ scenario[str(gemeinde_id)])
        
    if gemeinde_id==3:
        filename='input_data_g'+str(gemeinde_id)+'.xlsx'
        print('Gemeinde: '+ scenario[str(gemeinde_id)])
    
    if gemeinde_id == 4:
        filename= 'input_data_g'+str(gemeinde_id)+'.xlsx'
        print('Gemeinde: '+ scenario[str(gemeinde_id)])
    
    if gemeinde_id == 5:
        filename= 'input_data_g'+str(gemeinde_id)+'.xlsx'
        print('Gemeinde: '+ scenario[str(gemeinde_id)])
        
    return filename


# Function: Parse nodes data from excel sheets
def nodes_from_excel(filename):
    
    # Check: does Excel file exist?
    if not filename or not os.path.isfile(filename):
        raise FileNotFoundError('Excel data file {} not found.'
                                .format(filename))
    xls = pd.ExcelFile(filename)
  # Parse nodes data from excel sheets  
    if scenario_id == 0:
        buses_df = xls.parse('buses', index_col = 0)
        excess_df = xls.parse('excess', index_col = 0)
        shortage_df = xls.parse('shortage', index_col = 0)
        sources_df = xls.parse('sources', index_col = 0)
        renewables_df = xls.parse('renewables', index_col = 0)
        transformers_df = xls.parse('transformers', index_col = 0)
        demand_df = xls.parse('demand', index_col = 0)
        storages_df = xls.parse('storages', index_col = 0)   
        timeseries_df = xls.parse('time_series', parse_dates = True, index_col = 0)
    
    else:
        buses_df = xls.parse('buses', index_col = 0)
        excess_df = xls.parse('excess_'+str(scenario_id), index_col = 0)
        shortage_df = xls.parse('shortage_'+str(scenario_id), index_col = 0)
        sources_df = xls.parse('sources', index_col = 0)
        renewables_df = xls.parse('renewables_'+str(scenario_id), index_col = 0)
        transformers_df = xls.parse('transformers_'+str(scenario_id), index_col = 0)
        demand_df = xls.parse('demand_'+str(scenario_id), index_col = 0)
        storages_df = xls.parse('storages_'+str(scenario_id), index_col = 0)   
        timeseries_df = xls.parse('time_series_'+str(scenario_id), parse_dates = True, index_col = 0)
    
    
    return buses_df, sources_df,excess_df,shortage_df, renewables_df, transformers_df, demand_df,\
    storages_df, timeseries_df
    
# Function: Create nodes based on parsed data
    
def create_nodes(buses_df, sources_df, excess_df, shortage_df, renewables_df, transformers_df, demand_df, 
    storages_df, timeseries_df, starttime, endtime):
    
     node_dict={}
     # Create busses:
     for i, b in buses_df.iterrows():
         if b['active']:
            node_dict[i] = solph.Bus(label = i)
     
     # Create Excess :         
     # Input-Excel "exchange" = 1: variabel costs = time series
     # Especially for electricity Export: inverse day ahead price + transaction costs       
     for i, e in excess_df.iterrows():
         if e['active']:
              if e['exchange']:
                  node_dict[i] = solph.Sink(label=i, inputs={node_dict[e['from']]: solph.Flow(
                     variable_costs = np.array(timeseries_df[i][starttime:endtime]))})
              else:
                  node_dict[i] = solph.Sink(label=i, inputs={node_dict[e['from']]: solph.Flow(
                     variable_costs = e['excess_costs'])})
                  
     for i, s in shortage_df.iterrows():
         if s['active']:
             node_dict[i] = solph.Source(label=i, outputs={
             node_dict[s['to']]: solph.Flow(
                    variable_costs=s['shortage_costs'])})
             
     # Create sources: Input-Excel "exchange" = 1: variable costs = time series
     # Especially for electricity import: day ahead price
             
     for i, cs in sources_df.iterrows():
         if cs['active'] : 
             if cs['exchange']:
                 node_dict[i]= solph.Source(label=i, outputs={node_dict[cs['to']]:
                                            solph.Flow(
                                                #nominal_value= 200, DELETE
                                                #summed_max= cs['cap'],DELETE
                                                variable_costs = np.array(timeseries_df[i][starttime:endtime]),
                                                emission_factor = cs['emission_factor'])})
             else:
                node_dict[i] = solph.Source(label=i, outputs={node_dict[cs['to']]: 
                                            solph.Flow(
                                                #nominal_value= 1,DELETE
                                                #summed_max= cs['cap'],DELETE
                                                variable_costs = cs['variable_costs'] + cs['carbon_price'] * cs['emission_factor'],
                                                emission_factor = cs['emission_factor'])})
     
     # Create renewables(Wind and PV):
     # Input-Excel "exchange" = 1, Varible costs= timeseries
     # "exchange" not used due to assumption of fixed marginal costs               
     for i, re in renewables_df.iterrows():       
        if re['active']:
            if re['exchange']:
                 node_dict[i] = solph.Source(label=i, outputs = {node_dict[re['to']]: solph.Flow(
                            actual_value = np.array(timeseries_df[i][starttime:endtime]),
                            nominal_value=re['capacity'],
                            variable_costs = np.array(timeseries_df['EE_Preis'][starttime:endtime]),
                            emission_factor=re['emission_factor'],
                            fixed=True)})

            else:    
                node_dict[i] = solph.Source(label=i, outputs = {node_dict[re['to']]: solph.Flow(
                            actual_value = np.array(timeseries_df[i][starttime:endtime]),
                            nominal_value=re['capacity'],
                            emission_factor=re['emission_factor'],
                            fixed=True)})
     
        # Create sinks (as demands):  
        # for further variation an "fixed export" can be added
     for i, d in demand_df.iterrows():
        if d['active']:
            node_dict[i] = solph.Sink(label=i, inputs={node_dict[d['from']]: solph.Flow(
                            actual_value = np.array(timeseries_df[i][starttime:endtime]),
                            nominal_value=d['maximum']+d['fixed_export'],
                            fixed=True)})
    
    # Create transformers:
            
    # invest_mode = True: Transformers (PtH2, Netz (Im/Ex), Biomasse BHKW, Rückverstromung, Transport_Export_H2)
    # are defined by the Investment module. 
    # Additional Input-Excel: "investment" = 1 activates the investment mode
    # All transformers without "investment" can be limited by the cap variable.
    # Basic configuration: cap is big enough, no impact on Optimization, just needed for simplifieng the component generation
    # Cap is used for further variation scenarios.
    # invest_mode= False: all transfomrmers have a fixed capacity
   
     if invest_mode:
         for i, t in transformers_df.iterrows():
                 
            if t['active']:
               if t['investment']:
                  outflow_args_el = {
                      'investment': solph.Investment(
                                    maximum= t['potential_capacity'],
                                    ep_costs= t['opex']+ economics.annuity(capex= t['capex'],n=t['lifetime'],wacc=t['wacc']),
                                    minimum = t['min_capacity'],
                                    existing = t['capacity']),
                      'variable_costs': t['variable_input_costs']}
                  node_dict[i] = solph.Transformer(
                    label = i,
                    inputs = {node_dict[t['from']]: solph.Flow()},
                    outputs = {node_dict[t['to']]: solph.Flow(**outflow_args_el)},
                           conversion_factors = {node_dict[t['to']]: t['efficiency']})
                 
               else: 
                   if t['multi']:
                       outflow_args_el = {
                       'nominal_value': t['capacity']*0.5,
                       'summed_max': t['cap'],
                       'variable_costs': t['variable_input_costs']}
                       
                       outflow_args_th ={
                          'nominal_value': t['capacity']*0.5,
                          'summed_max': t['cap'],
                          'variable_costs': t['variable_input_costs']}
                       
                       node_dict[i] = solph.Transformer(
                           label = i,
                           inputs = {node_dict[t['from']]: solph.Flow()},
                           outputs = {node_dict[t['to']]: solph.Flow(**outflow_args_el),
                                      node_dict[t['to_multi']]: solph.Flow(**outflow_args_th)},
                           conversion_factors = {node_dict[t['to']]: t['efficiency'],
                                                 node_dict[t['to_multi']]: t['efficiency_multi']})
                       
                   else:
     # for further variation limitation of electricity import (autarky):
     # An annual limit is created:
     # summed_max * nominal_value = annual_limit  (for biomass and recovery summed_max sufficiently large)
                       outflow_args_el = {
                           'nominal_value': t['capacity'],
                           'summed_max': t['cap'],
                           'variable_costs': t['variable_input_costs']}
                       node_dict[i] = solph.Transformer(
                           label = i,
                           inputs = {node_dict[t['from']]: solph.Flow()},
                           outputs = {node_dict[t['to']]: solph.Flow(**outflow_args_el)},
                               conversion_factors = {node_dict[t['to']]: t['efficiency']})
     
     # invest_mode = False : create all transformers with a fixed capacity and without cap 
     else:
         for i, t in transformers_df.iterrows():
             if t['active']:
                outflow_args_el = {
                       'nominal_value': t['capacity'],
                       'variable_costs': t['variable_input_costs']}
                
                node_dict[i] = solph.Transformer(
                     label = i,
                     inputs = {node_dict[t['from']]: solph.Flow()},
                     outputs = {node_dict[t['to']]: solph.Flow(**outflow_args_el)},
                           conversion_factors = {node_dict[t['to']]: t['efficiency']})

     # create storages:  
     # "investment" =1 in Input-Excel: Activates the investment mode. Storages is set by investment module.
     # minimal capacity = 1 MW
     for i, st in storages_df.iterrows():
        if st['active']:
              if st['investment']:
               
                 outflow_args = {
                      'investment': solph.Investment(
                                    maximum = st['potential_capacity'],
                                    ep_costs= st['opex']+ economics.annuity(capex= st['capex'],n=st['lifetime'],wacc=st['wacc']),
                                    minimum = st['existing_storage_capacity'])}
               
                 node_dict[i] = solph.components.GenericStorage(
                    label = i,
                    inputs = {node_dict[st['bus']]: solph.Flow(
                        variable_costs = st['variable_input_costs'])},
                    outputs ={node_dict[st['bus']]: solph.Flow(
                        variable_costs = st['variable_output_costs'])},
                    loss_rate = st['loss_rate']/duration_year,
                    initial_storage_level = st['initial_storage_level'],
                    balanced = False,
                    max_storage_level = st['max_storage_level'],
                    min_storage_level = st['min_storage_level'],
                    inflow_conversion_factor = st['efficiency_inflow'],
                    outflow_conversion_factor = st['efficiency_outflow'],
                    invest_relation_input_capacity = st['invest_relation_input'],
                    invest_relation_output_capacity = st['invest_relation_output'],
                    **outflow_args)             
              
              # "investement"=0: storage with a fixed capacity  
              else:
                 node_dict[i] = solph.components.GenericStorage(
                    label = i,
                    nominal_storage_capacity = st['nominal_storage_capacity'],
                    inputs = {node_dict[st['bus']]: solph.Flow(
                        nominal_value = st['capacity_inflow'], 
                        variable_costs = st['variable_input_costs'])},
                    outputs ={node_dict[st['bus']]: solph.Flow(
                        nominal_value = st['capacity_outflow'], 
                        variable_costs = st['variable_output_costs'])},
                    loss_rate = st['loss_rate']/duration_year,
                    initial_storage_level = st['initial_storage_level'],
                    balanced = False, 
                    max_storage_level = st['max_storage_level'],
                    min_storage_level = st['min_storage_level'],
                    inflow_conversion_factor = st['efficiency_inflow'],
                    outflow_conversion_factor = st['efficiency_outflow'])

     return node_dict

# Initiate the logger
logger.define_logging(logfile='simple_dispatch_auto_v01.log',
                      screen_level=logging.INFO,
                      file_level=logging.DEBUG)
logging.info ('Initialize the energy system')

# Set datetime:
# starttime / endtime
duration_year = 8784
duration_time = 8784

starttime = '2016-01-01 00:00:00'
endtime = '2016-12-31 23:00:00'
year = 2016


# 2016 Leap year
datetime_index = pd.date_range(starttime, endtime,
                               freq='60min')

# Start creation and optimization
logging.info('Starting optimization')

# Initialise the energy system
es = solph.EnergySystem(timeindex=datetime_index)

# Define the modelled systems
logging.info('Definition Gemeinde')
# Choose modelled system
filename = define_gemeinde(gemeinde_id)

# show scenario
logging.info('Definition Szenario')
print('Szenario: '+ scenario[str(scenario_id)])

logging.info('Parse Nodes from excel sheet')

# Parse node date from excel sheets
buses_df, sources_df,excess_df, shortage_df, renewables_df, transformers_df, demand_df, storages_df, timeseries_df = nodes_from_excel(
    os.path.join(os.path.dirname(__file__), filename))


logging.info('Create nodes from excel sheet')

# Create nodes

node_dict = create_nodes (buses_df, sources_df, excess_df, shortage_df, renewables_df, transformers_df, demand_df, 
                          storages_df, timeseries_df, starttime, endtime)

# add nodes and flows to energy system
es.add(*node_dict.values())

print("*********************************************************")
print("The following objects have been created from excel sheet:")
for n in es.nodes:
    print(n)
# Show added nodes in console
print("*********************************************************")

# Create a least cost Optimization model from the energysystem
om = solph.Model(es)
 
# Calculate dual variable of the buses ! Actually: WARNING, when used but appeared as well with a other codes 
if obtain_duals:
    om.receive_duals()
    logging.info('Receive duals for price calculation')


logging.info('Solve the energysystem')

# Write an lp file with all target function and constraints
filename_lp = os.path.join(os.path.dirname(__file__),'pth2_v01.lp')

# solving the linear problem using the given solver
om.solve(solver='gurobi')

logging.info('Optimization successful')

###############################################################################
### Process Model results
###############################################################################

results = outputlib.processing.results(om)
meta_results = outputlib.processing.meta_results(om)

# Print Results
print(results)
print(meta_results)

# Store the results of the electricity/ hydrogen/ heat bus
results_el= outputlib.views.node(results, 'b_el')
results_el_im= outputlib.views.node(results,'b_el_im')
result_el_ex= outputlib.views.node(results, 'b_el_ex')
results_hydro= outputlib.views.node(results, 'b_hydro')
results_hydro_ex = outputlib.views.node(results,'b_hydro_ex')
results_hydro_storage = outputlib.views.node(results,'Wasserstoffspeicher')
results_th = outputlib.views.node(results,'b_th')

# store the results of the electricity bus, the heat bus and the hydrogen bus in a Dataframe
df_results_el = results_el['sequences']

df_results_el_im = results_el_im['sequences']

df_results_hydro = results_hydro['sequences']
df_results_hydro_ex = results_hydro_ex['sequences']
df_results_hydro_storage_level = results_hydro_storage['sequences'][(('Wasserstoffspeicher','None'),'capacity')]

df_results_th = results_th['sequences']

# gather Results for electricity and hydrogen in a DataFrame
df_results = df_results_el.join(df_results_hydro)


# Function: store results el/hydro in a DataFrame with an Label
def create_aggregated_el_results(results):
    
    # Define all Labels for sources/sinks around one bus
    # OderedDict: sorted Dictionary stores the order of entries
    energy_sources = ["PV","Wind", "Biomasse_BHKW","Netz_im","Rückverstromung"]
    energy_sources_dict = OrderedDict()
    
    energy_sinks = ["Stromnachfrage_HH","Stromnachfrage_GHD","Netz_ex","PtH2"]
    energy_sinks_dict = OrderedDict()
   
    # For-loop for In-Flows electricity bus: 
    for source in energy_sources:
        energy_sources_dict[source]= [entry for entry in results.columns.values if (source in entry [0][0] and "b_el" in entry [0][1] and "flow" in entry [1])]  
    
    # For-loop for out-flows electricity bus:
    for sink in energy_sinks:
        energy_sinks_dict[sink]= [entry for entry in results.columns.values if ("b_el" in entry [0][0] and sink in entry[0][1] and "flow" in entry [1])]    
    
    Production = pd.DataFrame()
    for key, val in energy_sources_dict.items():
        Production[key] = results[val].sum(axis =1)
        
    for key, val in energy_sinks_dict.items():
        Production[key] =results[val].sum(axis =1)
    
    return Production

# Function: Store results of import electricity bus with labels (similiar method like electricity bus)
def create_aggregated_el_im_results(results):
    
    import_sources = ["Strom_Import","b_el_im_shortage"]
    import_sources_dict = OrderedDict()
    
    import_sinks = ["Netz_im"]
    import_sinks_dict = OrderedDict()
    
    for source in import_sources: 
        import_sources_dict[source] = [entry for entry in results.columns.values if (source in entry [0][0] and "b_el_im" in entry [0][1] and "flow" in entry [1])]
    
    for sink in import_sinks:
        import_sinks_dict[sink]= [entry for entry in results.columns.values if ("b_el_im" in entry [0][0] and sink in entry[0][1] and "flow" in entry [1])]    
    
    Import = pd.DataFrame()
    for key, val in import_sources_dict.items():
        Import[key] = results[val].sum(axis =1)
        
    for key, val in import_sinks_dict.items():
        Import[key] =results[val].sum(axis =1)
    
    return Import

# Function: Store results of hydrogen bus in a DataFrame with labels    
def create_aggregated_hydro_results(results):
    
    hydro_sources= ["PtH2", "Wasserstoffspeicher","Import_H2"]
    hydro_sources_dict= OrderedDict()
    
    hydro_sinks = ["Nachfrage_H2","Wasserstoffspeicher","Transport_Export_H2","Rückverstromung"]
    hydro_sinks_dict = OrderedDict()
    
     # For-loop for In-Flows Hydrogen-Bus: 
    for source in hydro_sources:
        hydro_sources_dict[source]= [entry for entry in results.columns.values if (source in entry [0][0] and "b_hydro" in entry [0][1] and "flow" in entry [1])]

    # For-loop for Out-FLows Hydrogen-Bus: 
    for sink in hydro_sinks:
        # if-loop for Wasserstoffspeicher inflow, that Wasserstoffspeicher outflow is not overwritten
        if sink=="Wasserstoffspeicher":
            hydro_sinks_dict[sink+"_in"]= [entry for entry in results.columns.values if ("b_hydro" in entry [0][0] and sink in entry[0][1] and "flow" in entry [1])]
        else:
         hydro_sinks_dict[sink]= [entry for entry in results.columns.values if ("b_hydro" in entry [0][0] and sink in entry[0][1] and "flow" in entry [1])]
    
    # Gather Production results of source/sinks in one Dataframe
    Production = pd.DataFrame()
    for key, val in hydro_sources_dict.items():
        Production[key] = results[val].sum(axis=1)
        
    for key, val in hydro_sinks_dict.items():
        Production[key] = results[val].sum(axis=1)
    
    return Production

# Function: Store results for hydrogen export bus in a DataFrame with Labels
def create_aggregated_hydro_ex_results(results):
    
    import_sources = ["Transport_Export_H2"]
    import_sources_dict = OrderedDict()
    
    import_sinks = ["Export_H2"]
    import_sinks_dict = OrderedDict()
    
    for source in import_sources: 
        import_sources_dict[source] = [entry for entry in results.columns.values if (source in entry [0][0] and "b_hydro_ex" in entry [0][1] and "flow" in entry [1])]
    
    for sink in import_sinks:
        import_sinks_dict[sink]= [entry for entry in results.columns.values if ("b_hydro_ex" in entry [0][0] and sink in entry[0][1] and "flow" in entry [1])]    
    
    Export = pd.DataFrame()
    for key, val in import_sources_dict.items():
        Export[key] = results[val].sum(axis =1)
        
    for key, val in import_sinks_dict.items():
        Export[key] =results[val].sum(axis =1)
    
    return Export

# Function: Store results of heat bus in DataFrame with labels 
def create_aggregated_th_results(results):
    
    th_sources = ["Biomasse_BHKW"]
    th_sources_dict = OrderedDict()
    
    th_sinks = ["Export_Wärme"]
    th_sinks_dict = OrderedDict()
    
    for source in th_sources: 
        th_sources_dict[source] = [entry for entry in results.columns.values if (source in entry [0][0] and "b_th" in entry [0][1] and "flow" in entry [1])]
    
    for sink in th_sinks:
        th_sinks_dict[sink]= [entry for entry in results.columns.values if ("b_th" in entry [0][0] and sink in entry[0][1] and "flow" in entry [1])]    
    
    Production = pd.DataFrame()
    for key, val in th_sources_dict.items():
        Production[key] = results[val].sum(axis =1)
        
    for key, val in th_sinks_dict.items():
        Production[key] =results[val].sum(axis =1)
    
    return Production
    
# Function: Store dual variables for all busses in a DataFrame
def get_el_prices_from_duals(om, datetime_index):
    
    el_prices_list = []
    

    constr = [c for c in om.component_objects(po.Constraint, active = True) if c.name == "Bus.balance"][0]
    
    # duals for el
    el_prices_list = [om.dual[constr[index]] for index in constr if index[0].label == "b_el"]
    el_prices = pd.DataFrame(data= el_prices_list, index= datetime_index, columns = ["electricity price"])
    
    return el_prices

def get_el_im_prices_from_duals(om, datetime_index):
    
    el_im_prices_list = []
    

    constr = [c for c in om.component_objects(po.Constraint, active = True) if c.name == "Bus.balance"][0]
    
    # duals for el_im
    el_im_prices_list = [om.dual[constr[index]] for index in constr if index[0].label == "b_el_im"]
    el_im_prices = pd.DataFrame(data= el_im_prices_list, index= datetime_index, columns = ["electricity import price"])
    
    return el_im_prices

def get_hydro_prices_from_duals(om, datetime_index):
    
    hydro_prices_list = []    
        
    # Choose the Bus.balance constraint
    # (There is only one which is indexed by buses as well as timesteps)
    constr = [c for c in om.component_objects(po.Constraint, active = True) if c.name == "Bus.balance"][0]
    
    # Add the dual values of the hydrogen bus to the hydrogen prices list
    hydro_prices_list = [om.dual[constr[index]] for index in constr if index[0].label == "b_hydro"]
    hydro_prices = pd.DataFrame(data = hydro_prices_list, index = datetime_index, columns = ["hydrogen price"])
    
    return hydro_prices

# Store the dual variables of OM
if obtain_duals:
    hydro_prices = get_hydro_prices_from_duals(om, pd.date_range(starttime, endtime, freq = '60min'))
    el_prices = get_el_prices_from_duals(om, pd.date_range(starttime, endtime, freq = '60min'))
    el_im_prices = get_el_im_prices_from_duals(om, pd.date_range(starttime, endtime, freq = '60min'))
    
    logging.info('Prices are calculated')

# Function: Create automatically for listes buses overview for console
def create_table_cons_gen_for_all_busses(results_dict,scenario_id):
    for ele in results_dict:
        components_production = []
        components_demand =[]
        scenario_list=[scenario_id]
        
        
        if obtain_duals:
               results_dict[ele]['sequences'] = results_dict[ele]['sequences'].drop(columns=((ele,'None'),'duals'))
       
        for i in range(len(results_dict[ele]['sequences'].keys())):
            
            # Generation at bus
            if "(('"+ele in str(results_dict[ele]['sequences'].keys()[i]):
                replace_str= "(('"+ele+"', '"
                cut1 = str(results_dict[ele]['sequences'].keys()[i]).replace(replace_str,"")
                component = cut1.replace("'), 'flow')","")
                components_demand.append(component)

            # Consumption at bus
            if ele+"'), 'flow')" in str(results_dict[ele]['sequences'].keys()[i]):
                replace_str ="', '"+ele+"'), 'flow')"
                cut1 = str(results_dict[ele]['sequences'].keys()[i]).replace(replace_str,"")
                component = cut1.replace("(('","")
                components_production.append(component)
            
            
       # create table with results on component level:
        data_production = pd.DataFrame({
            'Generation':
                [results_dict[ele]['sequences'][((i,ele),'flow')].sum().round(3) for i in components_production],
            'Investment':
                [0 if 'scalars' not in results_dict[ele].keys() or ((i, ele), 'invest') not in results_dict[ele]['scalars'].index else results_dict[ele]['scalars'][((i, ele), 'invest')] for i in components_production],            
            'Scenario':
                scenario_list*len(components_production),
                },  index= [components_production])
        
   
        data_consumption =  pd.DataFrame({
            'Consumption':
                [results_dict[ele]['sequences'][((ele,i),'flow')].sum().round(3) for i in components_demand],
                 'Scenario':
                scenario_list*len(components_demand)
                },  index= [components_demand])
        
        
        
        #print results to consol for analysis    
        print('################################################################')
        print("Results for Bus: "+ele)    
        print(data_production)   
        print(data_consumption)

    return 

# Function: Store the results of OM all Busses
def create_bus_results(list_bus,results):
    results_dict={}
    for ele in list_bus:
        raw = outputlib.views.node(results, ele)
        results_dict.update({ele:raw})
    
    return results_dict
####### automatical Creation of results #################################
# List with busses to analyse results in console
list_bus = ["b_hydro","b_el","b_hydro_ex"]

# Create bus results by calling function create_bus_results with the bus_list
results_dict =  create_bus_results(list_bus,results)

# Create overview table for console analysis for choosen busses
table_results = create_table_cons_gen_for_all_busses(results_dict, scenario_id)

####### manual Creation of bus results #####################################
# Store results of OM
production_el = create_aggregated_el_results(df_results_el)
production_hydro = create_aggregated_hydro_results(df_results_hydro)
import_el = create_aggregated_el_im_results(df_results_el_im)
production_th = create_aggregated_th_results(df_results_th)
export_hydro = create_aggregated_hydro_ex_results(df_results_hydro_ex)

# Calculate residual load: (Stromnachfrage_HH + Stromnachfrage_GHD) - (PV+Wind)
residual_el = pd.DataFrame()
residual_el['Residuallast'] = (production_el['Stromnachfrage_HH']+ production_el['Stromnachfrage_GHD'])-(production_el['PV']+production_el['Wind'])

# Calculate residual load hydrogen: (Nachfrage_H2 - PtH2)
residual_hydro = pd.DataFrame()
residual_hydro['Residuallast'] = (production_hydro['Nachfrage_H2'])-(production_hydro['PtH2'])

# Store storage level:
hydro_storage_level = pd.DataFrame()
hydro_storage_level['Speicherstand'] = df_results_hydro_storage_level

logging.info('Results are aggregated')

# Calculate LCOH and system balance

def calculate_LCOH(production_el,production_hydro,production_th,timeseries_df,excess_df,shortage_df,sources_df, meta_results, results_hydro):
    
    # Dictionary for storing the calculations
    system_balance = {}
    
    income_el = (production_el['Netz_ex']*timeseries_df['b_el_ex_excess']).sum().round(2)
    income_hydro = (production_hydro['Transport_Export_H2']*excess_df['excess_costs']['Export_H2']).sum().round(2)
    income_th = (production_th['Biomasse_BHKW']*excess_df['excess_costs']['Export_Wärme']).sum().round(2)
    
    costs_hydro_im = (production_hydro['Import_H2']*shortage_df['shortage_costs']['Import_H2']).sum().round(2)
    system_costs = meta_results['objective']- income_el - income_hydro - income_th - costs_hydro_im
    
    # Calculate regional average weighted electricity price:
    # costs for imported electricity= hourly import* hourly electricity price
    el_costs = production_el['Netz_im']*timeseries_df['Strom_Import']
    
    # costs for power from biomass= hourly power generation from BHKW/ efficiency * price for biomass
    bio_costs = (production_el['Biomasse_BHKW']/0.45)*sources_df['variable_costs']['Biomasse_Import']
    
    # total electricty production from renewables, biomass, import
    production_el_sum = production_el['PV']+production_el['Wind']+production_el['Biomasse_BHKW']+production_el['Netz_im']
    
    # regional average electricity costs €/MWh
    average_el_costs = (el_costs+bio_costs)/(production_el_sum)
    
    # hydrogen costs = regional average electricity costs* electricity consumption of PtH2
    hydro_costs = average_el_costs*production_el['PtH2']
    
    # investment costs of PtH2 = capacity MW H2 * (annuity + OPEX)
    hydro_invest_costs= results_hydro['scalars'][(('PtH2', 'b_hydro'), 'invest')]*(economics.annuity(capex= transformers_df['capex']['PtH2'],n=transformers_df['lifetime']['PtH2'],wacc=transformers_df['wacc']['PtH2'])+transformers_df['opex']['PtH2']).round(2)
    
    # LCOH = hydrogen costs + investment cost of PtH2 / H2 production
    lcoh = (hydro_costs.sum()+hydro_invest_costs)/production_hydro['PtH2'].sum()
    
    # Store calculations in a Dataframe round 2 decimals
    system_balance = pd.DataFrame({
        'Objective': meta_results['objective'].round(2),
        'Einnahmen Stromexport': income_el.round(2),
        'Einnahmen Export Wärme': income_th.round(2),
        'Einnahmen Export H2': income_hydro.round(2),
        'Kosten Wasserstoffimporte': costs_hydro_im.round(2),
        'Kosten Stromimporte': el_costs.sum().round(2),
        'Systemkosten exkl. Einnahmen/H2-Importkosten':system_costs.round(2),
        'Kosten Strombedarf des Elektrolyseurs':hydro_costs.sum().round(2),
        'Investitionskosten des Elektrolyseurs':hydro_invest_costs.round(2),
        'LCOH 1/MWh': lcoh.round(2)}, index=["€"]).transpose()
    
    # Print LCOH in console
    print('###################################################################')
    print('Energiesystem Bilanz: ')
    print (system_balance)
    print('###################################################################')
    
    return system_balance

# Calculate LCOH and system balance by function
logging.info('Calculation of systemcosts and LCOH')
system_balance = calculate_LCOH(production_el,production_hydro,production_th,timeseries_df,excess_df,shortage_df,sources_df, meta_results,results_hydro)

# Create overview table for the export to excel
def create_overview_results(production_el, production_hydro, import_el, hydro_storage_level,transformers_df,renewables_df,storages_df):
    
    el_res_anlagen= ["PV", "Wind"]
    el_anlagen= ["Biomasse_BHKW","PtH2", "Rückverstromung"]
    hydro_anlagen = ["PtH2", "Wasserstoffspeicher","Wasserstoffspeicher_in"]
    el_netz = ["Netz_im","Netz_ex"]
    el_rest = ["Stromnachfrage_HH","Stromnachfrage_GHD"]
    hydro_rest = ["Import_H2","Transport_Export_H2","Nachfrage_H2"]

    
    el_res_anlagen_df = pd.DataFrame( {
                'Szenario': [scenario[str(scenario_id)]+" "+scenario[str(gemeinde_id)] for i in range(len(el_res_anlagen))],
                'Anlage': [el_res_anlagen[i] for i in range(len(el_res_anlagen))],
                'Produktion/Bedarf': [production_el[i].sum() for i in el_res_anlagen],
                'Kapazität': [renewables_df['capacity'][i] for i in el_res_anlagen],
                'Volllaststunden': [production_el[i].sum()/renewables_df['capacity'][i] if renewables_df['capacity'][i]>0 else 0 for i in el_res_anlagen],
                'Betriebsstunden':[(production_el[i]!=0).sum() for i in el_res_anlagen]})
    
    el_anlagen_df = pd.DataFrame( {
                'Szenario': [scenario[str(scenario_id)]+" "+scenario[str(gemeinde_id)] for i in range(len(el_anlagen))],
                'Anlage': [el_anlagen[i] for i in range(len(el_anlagen))],
                'Produktion/Bedarf': [production_el[i].sum() for i in el_anlagen],
                'Kapazität': [transformers_df['capacity'][i] if not transformers_df['investment'][i] else production_el[i].max() for i in el_anlagen],
                'Volllaststunden': [production_el[i].sum()/transformers_df['capacity'][i] if not transformers_df['investment'][i] or transformers_df['capacity'][i]<0 else production_el[i].sum()/production_el[i].max() if production_el[i].max()>0 else 0 for i in el_anlagen],
                'Betriebsstunden': [(production_el[i]!=0).sum() for i in el_anlagen]})
    
    concat1= pd.concat([el_res_anlagen_df,el_anlagen_df], ignore_index=True, sort=False)

    
    hydro_anlagen_df = pd.DataFrame( {
            'Szenario': [scenario[str(scenario_id)]+" "+scenario[str(gemeinde_id)] for i in range(len(hydro_anlagen))],
            'Anlage': [hydro_anlagen[i] for i in range(len(hydro_anlagen))],
            'Produktion/Bedarf': [production_hydro[i].sum() for i in hydro_anlagen],
            'Kapazität': [production_hydro[i].max() for i in hydro_anlagen],
            'Volllaststunden':[production_hydro[i].sum()/production_hydro[i].max() if production_hydro[i].max()>0 else 0 for i in hydro_anlagen],
            'Betriebsstunden': [(production_hydro[i]!=0).sum() for i in hydro_anlagen]})
    
    concat2= pd.concat([concat1,hydro_anlagen_df], ignore_index=True, sort=False)
    
    el_netz_df = pd.DataFrame({
            'Szenario': [scenario[str(scenario_id)]+" "+scenario[str(gemeinde_id)] for i in range(len(el_netz))],
            'Anlage': [el_netz[i] for i in range(len(el_netz))],
            'Produktion/Bedarf': [production_el[i].sum() for i in el_netz],
            'Kapazität':[transformers_df['capacity'][i] for i in el_netz],
            'Volllaststunden':[production_el[i].sum()/transformers_df['capacity'][i] for i in el_netz],
            'Betriebsstunden':[(production_el[i]!=0).sum() for i in el_netz]})
    
    concat3= pd.concat([concat2,el_netz_df], ignore_index=True, sort=False)
    
    el_rest_df = pd.DataFrame({
            'Szenario': [scenario[str(scenario_id)]+" "+scenario[str(gemeinde_id)] for i in range(len(el_rest))],
            'Anlage': [el_rest[i] for i in range(len(el_rest))],
            'Produktion/Bedarf': [production_el[i].sum() for i in el_rest],
            'Kapazität':["N.A." for i in el_rest],
            'Volllaststunden':["N.A." for i in el_rest],
            'Betriebsstunden':[(production_el[i]!=0).sum() for i in el_rest]})
    
    concat4= pd.concat([concat3,el_rest_df], ignore_index=True, sort=False)
    
    
    hydro_rest_df = pd.DataFrame({
            'Szenario': [scenario[str(scenario_id)]+" "+scenario[str(gemeinde_id)] for i in range(len(hydro_rest))],
            'Anlage': [hydro_rest[i] for i in range(len(hydro_rest))],
            'Produktion/Bedarf': [production_hydro[i].sum() for i in hydro_rest],
            'Kapazität':[production_hydro[i].max() for i in hydro_rest],
            'Volllaststunden':["N.A." for i in hydro_rest],
            'Betriebsstunden':[(production_hydro[i]!=0).sum() for i in hydro_rest]})
    
    concat5= pd.concat([concat4,hydro_rest_df], ignore_index=True, sort=False) 
    
    hydro_storage_df = pd.DataFrame({
             'Gesamte Speicherkapazität':[(hydro_storage_level['Speicherstand'].max()/storages_df['max_storage_level']['Wasserstoffspeicher'])]},
             index=["Wasserstoffspeicher", "Wasserstoffspeicher_in"])
    
    concat6 = concat5.join(hydro_storage_df, on="Anlage")                                    
    
    overview_results = concat6
    
    return overview_results

overview_results = create_overview_results(production_el, production_hydro, import_el, hydro_storage_level,transformers_df, renewables_df,storages_df)
logging.info('Overview results are created')


# Export Output of OM in Excel (Results_GEMEINDE.xls)
if export_output: 
    
    book = load_workbook('Results_'+scenario[str(gemeinde_id)]+'.xlsx')

    with pd.ExcelWriter('Results_'+scenario[str(gemeinde_id)]+'.xlsx', engine='openpyxl' ) as writer:
            writer.book = book
    # Convert the dataframe to an XlsxWriter Excel object.
            writer.sheets = dict((ws.title,ws) for ws in book.worksheets)
            overview_results.to_excel(writer, str(scenario_id)+"_overview_results")
            system_balance.to_excel(writer, str(scenario_id)+"_system_balance")
            production_el.to_excel(writer, str(scenario_id)+"_el_series")
            residual_el.to_excel(writer, str(scenario_id)+"_el_series", startcol=10)
            production_hydro.to_excel(writer,str(scenario_id)+"_hydro_series")
            residual_hydro.to_excel(writer, str(scenario_id)+"_hydro_series",startcol=8)
            hydro_storage_level.to_excel(writer,str(scenario_id)+"_hydro_series", startcol=10)
            import_el.to_excel(writer, str(scenario_id)+"_el_im_series")
            export_hydro.to_excel(writer, str(scenario_id)+"_hydro_ex_series")
            production_th.to_excel(writer, str(scenario_id)+"_th_series")
            
    
    # Duals are obtained, convert them to a dataframe:      
            if obtain_duals:
                timeseries_df['Strom_Import'].to_excel(writer, str(scenario_id)+"_prices", startcol=0)
                el_prices.to_excel(writer, str(scenario_id)+"_prices",startcol=2)
                hydro_prices.to_excel(writer,str(scenario_id)+"_prices",startcol=4)
                el_im_prices.to_excel(writer, str(scenario_id)+"_prices", startcol=6)
            

    # Close the Pandas Excel writer and output the Excel file.
            writer.save()
            logging.info('Results are exported to Excel')


print(df_results_el.sum())
print(df_results_hydro.sum())

logging.info('Start plotting')
###############################################################################
### Plot the results
###############################################################################
logging.info('Print the results')
# Print Results

# Electricity price 
font = {'family' : 'Arial',
        'weight' : 'bold',
        'size'   : 20}

plt.rc('font', **font)


plt.figure(figsize = (30,15))
plt.xlabel('time')
plt.ylabel('€/MWh')
plt.title('Börsenstrompreis')
plt.plot(timeseries_df['Strom_Import'], linewidth = 2.0)

###############################################################################
### Plots for residual load el/hydro
###############################################################################
plt.style.use('ggplot')
plt.figure(figsize = (20,5))

plt.xlabel('time')
plt.ylabel('MWh')
plt.title('Residuallast '+scenario[str(gemeinde_id)]+' '+scenario[str(scenario_id)])
residual_el_sorted= residual_el.sort_values('Residuallast', ascending=False,inplace=False)
reset_residual_el_sorted=residual_el_sorted.reset_index(drop=True, inplace=False)
plt.plot(reset_residual_el_sorted, linewidth = 2.0)

if save_plots:
    path= "./results/"+scenario[str(gemeinde_id)]+"/"
    plt.savefig(path+scenario[str(scenario_id)]+'_residuallast_el.png', dpi = 150, bbox_inches="tight")

plt.figure(figsize = (20,5))
plt.xlabel('time')
plt.ylabel('MWh')
plt.title('Residuallast H2 '+scenario[str(gemeinde_id)]+' '+scenario[str(scenario_id)])
residual_hydro_sorted= residual_hydro.sort_values('Residuallast', ascending=False,inplace=False)
reset_residual_hydro_sorted=residual_hydro_sorted.reset_index(drop=True, inplace=False)
plt.plot(reset_residual_hydro_sorted, linewidth = 2.0)

if save_plots:
    path= "./results/"+scenario[str(gemeinde_id)]+"/"
    plt.savefig(path+scenario[str(scenario_id)]+'_residuallast_hydro.png', dpi = 150, bbox_inches="tight")

###############################################################################
### Plots for electricity import/export
###############################################################################
plt.style.use('ggplot')

x= datetime_index
line_im = production_el['Netz_im']
line_ex = production_el['Netz_ex']

fig8, ax= plt.subplots(1,figsize=(30,15))
fig8.suptitle('Stromimport und -export '+scenario[str(gemeinde_id)]+' '+scenario[str(scenario_id)], size=20)
ax.set_xlabel('time')
ax.set_ylabel('MWh')
ax.plot(x,line_im, color='grey', label='Netz_im')
ax.plot(x,line_ex, color='darkred', label='Netz_ex')
ax.set_ylim(ymax=100,ymin=-50)

ax2= ax.twinx()
ax2.set_ylabel('€/MWh')
line_price= timeseries_df['Strom_Import']
ax2.plot(line_price, color="steelblue", label='Börsenstrompreis')

def align_yaxis(ax1, v1, ax2, v2):
    # adjust ax2 ylimit so that v2 in ax2 is aligned to v1 in ax1
    _, y1 = ax1.transData.transform((0, v1))
    _, y2 = ax2.transData.transform((0, v2))
    inv = ax2.transData.inverted()
    _, dy = inv.transform((0, 0)) - inv.transform((0, y1-y2))
    miny, maxy = ax2.get_ylim()
    ax2.set_ylim(miny+dy, maxy+dy)

#ax.plot(range(6),(8,-5.2,2,0,0,0))

align_yaxis(ax,0,ax2,0)

ax.legend(loc="upper left",fontsize=20)
ax2.legend(loc='upper right', fontsize=20)

fig8.tight_layout()

if save_plots:
    path= "./results/"+scenario[str(gemeinde_id)]+"/"
    plt.savefig(path+scenario[str(scenario_id)]+'_stromimport_export.png', dpi = 150, bbox_inches="tight")

plt.show()

###############################################################################
### Stackplot Strombus
###############################################################################
labels_el = production_el.columns.values
y1= [production_el[el] for el in labels_el]
#y1 = y1[:-1]

# Define color dictionary
colors = VisioData_auto_v01.set_color_settings().values()

_ = plt.style.use('ggplot')

fig3, ax3 =plt.subplots(1,1,figsize =(20,10), sharex=True)
fig3.suptitle('Strom Bus Stackplot')
fig3.subplots_adjust(top=2.5)

_ = ax3.stackplot(production_el.index, y1, colors=colors, labels=labels_el)
_ = ax3.legend(bbox_to_anchor = [1.1, 0.05, 0, 1])
_ = plt.axis(xmin = production_el.index[0], xmax = production_el.index[-1])
_ = plt.xlabel('time')
_ = plt.xticks(rotation=45)
_ = plt.ylabel('capacity in MW')
#    _ = ax1.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d %H:%M'))
_ = plt.tight_layout()

if save_plots:
    path= "./results/"+scenario[str(gemeinde_id)]+"/"
    plt.savefig(path+scenario[str(scenario_id)]+'_Strom_Stackplot.png', dpi = 150, bbox_inches="tight")

plt.show()

###############################################################################
### Stackplot Hydrogen bus
###############################################################################
labels_hydro = production_hydro.columns.values
y1= [production_hydro[el] for el in labels_hydro]
#y1 = y1[:-1]

colors = VisioData_auto_v01.set_color_settings_hydro().values()

_ = plt.style.use('ggplot')

fig4, ax4 =plt.subplots(1,1,figsize =(20,10), sharex=True)
fig4.suptitle('Wasserstoff Bus Stackplot')
fig4.subplots_adjust(top=2.5)

_ =ax4.stackplot(production_hydro.index, y1, colors=colors, labels=labels_hydro)
#_ = ax3.plot(production_hydro['Nachfrage_H2'], linestyle = 'dashed', color = 'black', label = 'demand')
_ = ax4.legend(bbox_to_anchor = [1.1, 0.05, 0, 1],loc='center right')
_ = plt.axis(xmin = production_hydro.index[0], xmax = production_hydro.index[-1])
_ = plt.xlabel('time')
_ = plt.xticks(rotation=45)
_ = plt.ylabel('capacity in MW')
#    _ = ax1.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d %H:%M'))
_ = plt.tight_layout()

if save_plots:
    path= "./results/"+scenario[str(gemeinde_id)]+"/"
    plt.savefig(path+scenario[str(scenario_id)]+'_Wasserstoff_Stackplot.png', dpi = 150, bbox_inches="tight")

###############################################################################
### Plot for hydrogen dual variable
###############################################################################
if obtain_duals:
    fig5, ax5 =plt.subplots(1,1,figsize =(20,10), sharex=True)
    fig5.suptitle('Wasserstoff Preis')
    fig5.subplots_adjust(top=2.5)

    _ = ax5.plot(hydro_prices.index, hydro_prices['hydrogen price'],label='hydrogen price')
    _ = ax5.legend(bbox_to_anchor = [1.1, 0.05, 0, 1],loc = 'center right')
    _ = plt.axis(xmin = hydro_prices.index[0], xmax = hydro_prices.index[-1])
    _ = plt.xlabel('time')
    _ = plt.xticks(rotation=45)
    _ = plt.ylabel('Preis in €/MW')
    #    _ = ax1.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d %H:%M'))
    _ = plt.tight_layout()


    if save_plots:
        path= "./results/"+scenario[str(gemeinde_id)]+"/"
        plt.savefig(path+scenario[str(scenario_id)]+'Hydrogen_dual_price.png', dpi = 150, bbox_inches="tight")
    
plt.show()

###############################################################################
### Plots with Visio
###############################################################################
### Color Dictionary
cdict = {
    (('b_el', 'Stromnachfrage_HH'), 'flow'): 'pink',
    (('b_el', 'Stromnachfrage_GHD'), 'flow'): 'red',
    #(('b_el', 'b_el_excess'), 'flow'): 'black',
    (('b_el','Netz_ex'),'flow'): 'darkgrey',
    (('b_el','PtH2'),'flow'):'lightblue',
    (('PV', 'b_el'), 'flow'): 'yellow',
    (('Wind', 'b_el'), 'flow'): 'blue',
    (('Netz_im', 'b_el'), 'flow'): 'lightgrey',
    (('Biomasse_BHKW','b_el'),'flow'):'green',
    (('Rückverstromung','b_el'),'flow'):'purple',
    (('b_hydro','Wasserstoffspeicher'),'flow'):'darkblue',
    (('b_hydro','Transport_Export_H2'),'flow'):'black',
    (('b_hydro','Rückverstromung'),'flow'):'purple',
    (('b_hydro','Nachfrage_H2'),'flow'):'greenblue',
    (('PtH2','b_hydro'),'flow'):'lightblue',
    (('Import_H2','b_hydro'),'flow'):'darkgrey',
    (('Wasserstoffspeicher','b_hydro'),'flow'):'blue'
    }

###############################################################################
### Smooth Plot Electricity
# inorder : list order of columns to plot the line plot , outorder order of columns to plot the bar plot
inorder_el = [(('Wind', 'b_el'), 'flow'),
           (('Biomasse_BHKW', 'b_el'), 'flow'),
           (('Rückverstromung','b_el'),'flow'),
           (('PV', 'b_el'), 'flow'),
           (('Netz_im', 'b_el'), 'flow')]

# Excess not autimatically added to legend, instead the grid ex is used
outorder_el =[(('b_el','Stromnachfrage_HH'),'flow'),
              (('b_el','Stromnachfrage_GHD'),'flow'),
              (('b_el','PtH2'),'flow'),
              (('b_el','Netz_ex'),'flow')
              ]

fig = plt.figure(figsize=(30, 15))

#plot_slice= oev.plot.slice_df(df_results_el, date_from = pd.datetime(year, 1,1),date_to= pd.datetime(year,12,31))


# Plotting a combined bar and line plot of a bus to see the fitting of in- and out-coming flows of the bus balance

#el_plot1 = oev.plot.io_plot('b_el',plot_slice, cdict= cdict, inorder=inorder_el, outorder=outorder_el,
#                            ax=fig.add_subplot(1,1,1), smooth= True)

ax= VisioData_auto_v01.shape_legend('b_el',**el_plot1)

#ax= oev.plot.set_datetime_ticks(ax, plot_slice.index, number_autoticks=12, date_format= '%d-%m')
ax.set_ylabel('MWh')
ax.set_xlabel('Zeit')
ax.set_title(scenario[str(scenario_id)]+' '+scenario[str(gemeinde_id)])

if save_plots:
    path= "./results/"+scenario[str(gemeinde_id)]+"/"
    plt.savefig(path+scenario[str(scenario_id)]+'_Strom_sliced.png', dpi = 150, bbox_inches="tight")

plt.show()

# Smooth Plot Hydrogen not used 
'''inorder_hydro = [(('PtH2', 'b_hydro'), 'flow'),
            (('Wasserstoffspeicher', 'b_hydro'), 'flow'),
           (('Import_H2','b_hydro'),'flow')]

outorder_hydro =[(('b_hydro','Nachfrage_H2'),'flow'),
              (('b_hydro','Wasserstoffspeicher'),'flow'),
              (('b_hydro','Rückverstromung'),'flow')
              ]

fig_hydro = plt.figure(figsize=(30, 15))

plot_slice= oev.plot.slice_df(df_results_hydro, date_from = pd.datetime(year, 11, 1),date_to= pd.datetime(year,12,31))

hydro_plot1 = oev.plot.io_plot('b_hydro',plot_slice, cdict= cdict, inorder=inorder_hydro,outorder=outorder_hydro, ax=fig_hydro.add_subplot(1,1,1),
                           smooth= True)

ax= VisioData_auto_v01.shape_legend('b_hydro',**hydro_plot1)

ax= oev.plot.set_datetime_ticks(ax, plot_slice.index, number_autoticks=12, date_format= '%d-%m')
ax.set_ylabel('MWh')
ax.set_xlabel('Zeit')
ax.set_title(scenario[str(scenario_id)]+' '+scenario[str(gemeinde_id)])


if save_plots:
    path= "./results/"+scenario[str(gemeinde_id)]+"/"
    plt.savefig(path+scenario[str(scenario_id)]+'_Wasserstoff_sliced.png', dpi = 150, bbox_inches="tight")
'''

plt.show()

# Histogram Electricity price

n_bins = 50
fig_hist,axs = plt.subplots(1,1, sharey=True, tight_layout=True)

# We can set the number of bins with the `bins` kwarg
axs.hist(timeseries_df['Strom_Import'], bins=n_bins)

plt.show()

logging.info('DONE')